from typing import Optional
from pyexlatex.models.item import SimpleItem


class Text(SimpleItem):
    """
    Gets back to raw text, useful to have normal text formatting inside of equations.
    """
    name = 'text'

    def __init__(self, env: str, modifiers: Optional[str] = None):
        self.env = env
        super().__init__(self.name, env, modifiers)



from typing import Optional, Dict, Any, List, Tuple
from pyexlatex.logic.pdf.errors.exc import (
    TooManyUnprocessedFloatsException,
    OutputLoopConsecutiveDeadCycles,
    LatexException
)
from pyexlatex.logic.pdf.api.exc_handler.prepend.floats import get_extra_float_and_new_num_floats
from pyexlatex.logic.pdf.api.exc_handler.prepend.cycles import get_max_dead_cycles_and_new_num_cycles
from pyexlatex.logic.pdf.api.exc_handler.prepend.typing import PrependItemsDict, PrependKwargsDict


def handle_prepend_exceptions(exceptions: List[LatexException], prepend_kwarg_dict: PrependKwargsDict = None,
                              prepend_items_dict: PrependItemsDict = None
                              ) -> Tuple[PrependItemsDict, PrependKwargsDict, List[LatexException]]:
    if prepend_items_dict is None:
        prepend_items_dict = {}
    unhandled_exceptions = []
    for exception in exceptions:
        try:
            prepend_item, prepend_kwarg_dict = get_data_to_handle_prepend_exception(exception, prepend_kwarg_dict)
            # Assigning to class as key ensures that there will only be one item of each type
            prepend_items_dict[type(prepend_item)] = prepend_item
        except LatexException as e:
            unhandled_exceptions.append(e)
    return prepend_items_dict, prepend_kwarg_dict, unhandled_exceptions


def get_data_to_handle_prepend_exception(exception: LatexException, prepend_kwarg_dict: PrependKwargsDict = None
                                         ) -> Tuple[Any, PrependKwargsDict]:

    prepend_kwarg_dict = get_prepend_kwarg_dict(prepend_kwarg_dict)

    try:
        raise exception
    except TooManyUnprocessedFloatsException:
        extra_float, new_num_floats = get_extra_float_and_new_num_floats(prepend_kwarg_dict['extra_floats_num'])
        prepend_kwarg_dict['extra_floats_num'] = new_num_floats
        return extra_float, prepend_kwarg_dict
    except OutputLoopConsecutiveDeadCycles:
        max_dead_cycles, new_num_cycles = get_max_dead_cycles_and_new_num_cycles(prepend_kwarg_dict['cycles_num'])
        prepend_kwarg_dict['cycles_num'] = new_num_cycles
        return max_dead_cycles, prepend_kwarg_dict


def get_prepend_kwarg_dict(prepend_kwarg_dict: PrependKwargsDict = None) -> Dict[str, Any]:
    if prepend_kwarg_dict is None:
        prepend_kwarg_dict = {}

    required_keys = [
        'extra_floats_num',
        'cycles_num'
    ]

    for key in required_keys:
        if key not in prepend_kwarg_dict:
            prepend_kwarg_dict[key] = None

    return prepend_kwarg_dict


def add_prepend_items_dict_to_latex_str(prepend_items_dict: PrependItemsDict, latex_str: str) -> str:
    from pyexlatex.logic.builder import _build
    if not prepend_items_dict:
        return latex_str
    prepend_items = [str(item) for item in prepend_items_dict.values()]
    return _build(prepend_items + [latex_str])


from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.item import ItemBase


class Template(TextAreaMixin, ItemBase):
    r"""
    Base class for creating custom Pyexlatex templates.

    To use, simply subclass Template, then in __init__, set the attribute .contents, and call super().__init__().
    The .contents attribute should be set to a string, pyexlatex object, or iterable of those.

    Examples:

    Contents can be a Pyexlatex object or string:

    >>> import pyexlatex as pl
    >>> class BoldUnderline(pl.Template):
    >>>
    >>>     def __init__(self, contents):
    >>>         self.contents = pl.Bold(pl.Underline(contents))
    >>>         super().__init__()
    >>>
    >>> print(BoldUnderline('woo'))
    \textbf{\underline{woo}}

    Contents can also be an iterable of strings or Pyexlatex objects

    >>> import pyexlatex as pl
    >>> class BoldAndUnderline(pl.Template):
    >>>
    >>>     def __init__(self, bold_contents, underline_contents, *other_contents):
    >>>         self.contents = [
    >>>             pl.Bold(bold_contents),
    >>>             pl.Underline(underline_contents),
    >>>             *other_contents
    >>>         ]
    >>>         super().__init__()
    >>>
    >>> print(BoldAndUnderline('woo bold', 'woo underline', 'something'))
    \textbf{woo bold}
    \underline{woo underline}
    something

    """

    def __init__(self):
        self._validate_contents()
        super().__init__(None, self.contents)

    def __str__(self):
        if isinstance(self.contents, str):
            return self.contents
        elif hasattr(self.contents, 'is_LatexItem') and self.contents.is_LatexItem:
            # Another pyexlatex item is the root contents, allow it to determine str output
            return str(self.contents)
        else:
            # An iterable of pyexlatex items
            from pyexlatex.logic.builder import _build
            return _build(self.contents)

    def _validate_contents(self):
        contents_exception = ContentParseException(
            f'not able to parse attribute .contents in {self.__class__.__name__}. '
            f'Ensure that .contents is set to a string, a pyexlatex item, or '
            f'an iterable. Got {self.contents} of type {type(self.contents)}.'
        )

        if isinstance(self.contents, str):
            return
        elif hasattr(self.contents, 'is_LatexItem') and self.contents.is_LatexItem:
            return
        elif isinstance(self.contents, dict):
            raise contents_exception
        else:
            try:
                from pyexlatex.logic.builder import _build
                _build(self.contents)
            except TypeError as e:
                if 'can only join an iterable' in e:
                    raise contents_exception



class ContentParseException(Exception):
    pass




from typing import Optional, Any
from pyexlatex.models.item import ItemBase
from pyexlatex.texgen import no_options_no_contents_str
from mixins.repr import ReprMixin

class Mode(ItemBase, ReprMixin):
    name = 'mode'
    repr_cols = [
        'mode_type',
        'contents'
    ]

    def __init__(self, mode_type: str = 'presentation', contents: Optional[Any] = None):
        self.mode_type = mode_type
        self.contents = contents
        super().__init__()

    def __str__(self):
        """
        \mode must be on its own line with no whitespace in some cases, so must have a custom __str__ method
        """
        from pyexlatex.logic.builder import _build
        from pyexlatex.logic.format.contents import format_contents
        mode_definition = no_options_no_contents_str(self.name)  # \mode
        mode_type_str = f'<{self.mode_type}>'
        contents_str = f'{{{format_contents(self.contents)}}}'
        mode_contents = f'{mode_type_str}{contents_str}'
        return _build([
            mode_definition,
            mode_contents
        ])




from typing import Sequence, List, Union
from pyexlatex.models.format.join import Joined


def join_with_commas_and_and(items: Sequence[str]) -> str:
    if len(items) == 0:
        return ''
    elif len(items) == 1:
        return items[0]
    elif len(items) == 2:
        # Use only and, not commas
        return f'{items[0]} and {items[1]}'

    # Length 3+, handled the same, commas for all and also and on the last join
    out_str = ', '.join(items[:-1])
    out_str += f', and {items[-1]}'
    return out_str


def join_with_commas_and_and_output_list(items: Sequence[str]) -> List[str]:
    if len(items) == 0:
        return ['']
    elif len(items) == 1:
        return [items[0]]
    elif len(items) == 2:
        # Use only and, not commas
        return [items[0], ' and ', items[1]]

    # Length 3+, handled the same, commas for all and also and on the last join
    out_list: List[Union[str, Joined]] = []
    for item in items[:-1]:
        out_list.append(Joined([item, ', ']))
    out_list[-1] = Joined([items[-2], ', and '])  # remove comma after last item added, replace with comma and and
    out_list.append(items[-1])

    return out_list



from pyexlatex.models.item import SimpleItem


class Value(SimpleItem):
    name = 'value'

    def __init__(self, of: str, **kwargs):
        self.of = of

        super().__init__(self.name, self.of, **kwargs)



from mixins.repr import ReprMixin
from pyexlatex.table.models.mixins.addvalues.row import RowAddMixin
from pyexlatex.table.models.labels.multicolumnlabel import MultiColumnLabel
from pyexlatex.texgen.replacements.file import general_latex_replacements


class Label(ReprMixin, RowAddMixin):
    """
    Only necessary to use if need to set span or alignment manually. Can construct LabelCollection objects from Labels.

    Useful for constructing custom multicolumn labels.

    >>>import pyexlatex.table as lt
    >>>data_table = lt.DataTable.from_df(some_df)
    >>>label = lt.Label('Long label', span=5, align='r')
    >>>long_label_collection = lt.LabelCollection([label], underline='1-5')
    >>>
    >>>data_table.column_labels.insert(long_label_collection, 0)
    """
    repr_cols = ['value', 'span']

    def __init__(self, value, span: int=1, align='c'):
        self.value = value
        self.span = span
        self.align = _set_align(align)

    def __eq__(self, other):
        if hasattr(other, 'value'):
            return self.value == other.value
        else:
            return self.value == other

    def __str__(self):
        latex_valid_value = general_latex_replacements(str(self.value))
        if len(self) == 1:
            return str(latex_valid_value)
        else:
            return str(MultiColumnLabel(latex_valid_value, span=self.span, align=self.align, add_table_line_break=False))

    def __len__(self):
        return self.span

    def _add_class(self, other):
        from pyexlatex.table.models.table.row import Row
        from pyexlatex.table.models.labels.row import LabelRow
        # keep same class if both are same class
        # otherwise, default to Row class
        self_class = type(self)
        other_class = type(other)
        klass = LabelRow if self_class == other_class else Row

        return klass

def _set_align(align: str):
    align = align.lower().strip()

    assert align in ('l', 'c', 'r')

    return align


from typing import Union, Sequence
from pyexlatex.presentation.beamer.overlay.overlay_param import OverlayParameter

OverlayOption = Union[OverlayParameter, int]


class Overlay:
    """
    Creates an effect on a frame such as appearing on a later frame.
    """

    def __init__(self, overlay_options: Sequence[OverlayOption]):
        self.overlay_options = overlay_options

    @property
    def overlay_options_str(self) -> str:
        return ','.join([str(option) for option in self.overlay_options])

    def __str__(self) -> str:
        if self.overlay_options_str == '':
            return ''
        return f'<{self.overlay_options_str}>'


from pyexlatex.models.item import EqualsItem

class MaxDeadCycles(EqualsItem):
    name = 'maxdeadcycles'

    def __init__(self, num_cycles: int = 200):
        super().__init__(self.name, str(num_cycles))




from pyexlatex.models.package import Package
from pyexlatex.table.models.texgen.newcolumntype import NewColumnTypes

class ColumnTypesPackage(Package):
    """
    New column types are created by importing the dcolumn package then adding the column type definition lines
    """

    def __init__(self):
        super().__init__('dcolumn')

    def __str__(self):
        from pyexlatex.logic.builder import _build
        package_import = super().__str__()
        return _build([
            package_import,
            NewColumnTypes()
        ])



from pyexlatex.models.item import NoOptionsNoContentsItem


class RaggedRight(NoOptionsNoContentsItem):
    name = 'raggedright'

    def __init__(self, **kwargs):
        super().__init__(self.name, **kwargs)


from typing import Optional
from pyexlatex.models.item import (
    StringAdditionMixin,
    IsSpecificClassMixin,
    IsLatexItemMixin,
    _basic_item_str,
    _multi_option_item_str,
    ItemBase
)

class LineSpacing(ItemBase):

    def __init__(self, line_spacing: float):
        self.logical_line_spacing = line_spacing
        self.latex_line_spacing = latex_line_spacing_from_logical_line_spacing(line_spacing)
        super().__init__()

    @property
    def name(self):
        if self.logical_line_spacing == 2:
            return 'doublespacing'
        if self.logical_line_spacing == 1.5:
            return 'onehalfspacing'
        if self.logical_line_spacing == 1:
            return 'singlespacing'
        else:
            return 'setstretch'

    @property
    def _options_str(self) -> str:
        if self.logical_line_spacing in (1, 1.5, 2):
            return ''
        return f'{self.latex_line_spacing}'

    def __str__(self):
        options = self._options_str
        if options:
            return _basic_item_str(self.name, options)
        return _multi_option_item_str(self.name)


def latex_line_spacing_from_logical_line_spacing(line_spacing: float) -> float:
    """
    Latex for some reason has 1.65 as double line spacing, 1.325 as one and a half
    line spacing, and 1 as single spacing. Take an input on a normal scale (2 is
    double spaced, 1 is single space, 1.5 is one and a half spacing, and so on), and
    convert to the latex scale.
    Args:
        line_spacing:

    Returns:

    """
    return round((line_spacing - 1) * (0.65/1) + 1, 2)



from pyexlatex.models.item import MultiOptionSimpleItem
from pyexlatex.models.section.base import TextAreaMixin


class SetBeamerTemplate(TextAreaMixin, MultiOptionSimpleItem):
    name = 'setbeamertemplate'

    def __init__(self, beamer_item: str, content, **kwargs):
        TextAreaMixin.__init__(self, self.name, content)
        MultiOptionSimpleItem.__init__(self, self.name, beamer_item, self.contents, **kwargs)



from pyexlatex.models.item import NoOptionsNoContentsItem


class PageBreak(NoOptionsNoContentsItem):
    """
    Adds a break point where a new page should be started.
    """
    name = 'newpage'

    def __init__(self):
        super().__init__(self.name)



from pyexlatex.models.item import SimpleItem, ItemBase
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.logic.builder import _build


class RightHeader(TextAreaMixin, SimpleItem):
    name = 'rhead'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class CenterHeader(TextAreaMixin, SimpleItem):
    name = 'chead'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class LeftHeader(TextAreaMixin, SimpleItem):
    name = 'lhead'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class Header(ContainerItem, ItemBase):
    """
    Adds a header to a document, aligned to right, center, or left.
    """

    def __init__(self, contents, align: str = 'right'):
        """

        :param contents:
        :param align: one of 'right', 'left', 'center'
        """
        self.align = align.lower()
        self.contents = self._get_positioned_header(contents)
        super().__init__(self.contents.name, self.contents)

    def _get_positioned_header(self, contents):
        if self.align == 'right':
            return RightHeader(contents)
        elif self.align == 'left':
            return LeftHeader(contents)
        elif self.align == 'center':
            return CenterHeader(contents)
        else:
            raise ValueError('must pass one of "right", "left", "center" for align')

    def __str__(self):
        return str(self.contents)


class FancyHeader(SimpleItem):
    name = 'fancyhead'

    def __init__(self, contents):
        super().__init__(self.name, contents)

remove_header_line = r'\renewcommand{\headrulewidth}{0pt}'
remove_header = _build([
    remove_header_line,
    FancyHeader('')  # remove header default contents
])


from typing import Dict, List

latex_escape_begin_character = '\\' # literal \
latex_escape_end_characters = '{}'
latex_replacement_no_end_character_items = ['%', '&', '_', '#', '^']
latex_replacement_with_end_character_items = ['^']
latex_block_escape_characters = ['$']
replacement_dict = {
        latex_replacement_item: latex_escape_begin_character + latex_replacement_item
        for latex_replacement_item in latex_replacement_no_end_character_items
    }
replacement_dict.update({
        latex_replacement_item: latex_escape_begin_character + latex_replacement_item + latex_escape_end_characters
        for latex_replacement_item in latex_replacement_with_end_character_items
})

class Replacer:
    _first_letters: List[str]
    _unique_full_first_letter_replacements: Dict[str, str]

    def __init__(self, replacement_dict: Dict[str, str], escape_next_character_characters: List[str]=None,
                 escape_until_same_next_character_characters: List[str]=None):
        """

        Args:
            replacement_dict: dict where keys are strings to be replaced, and values are replacements
            escape_next_character_characters: list of one character strings for which after that character,
                the next character should not be replaced
            escape_until_same_next_character_characters: list of one character strings for which after that
                character, until this same character is seen again, characters should not be replaced
        """
        self.replacement_dict = replacement_dict

        if escape_next_character_characters is None:
            self.escape_next_character_characters = []
        else:
            self.escape_next_character_characters = escape_next_character_characters

        if escape_until_same_next_character_characters is None:
            self.escape_until_same_next_character_characters = []
        else:
            self.escape_until_same_next_character_characters = escape_until_same_next_character_characters

    def replace(self, string: str) -> str:
        possible_replacement = ''
        replacing = False
        output_str = ''
        previous_letter = ''
        escaped_until_character = ''  # blank string when not in an escaped block
        for letter in string:
            letter_is_escaped = previous_letter in self.escape_next_character_characters
            wipe_replacement = True # wipe unless told not to by the loop
            output_letter = letter
            possible_replacement += letter

            # Check if we're in an escaped block
            if escaped_until_character != '':
                if possible_replacement == escaped_until_character and not letter_is_escaped:
                    # End of escaped block
                    escaped_until_character = ''
            elif possible_replacement in self.escape_until_same_next_character_characters and not letter_is_escaped:
                # escaped block is starting
                escaped_until_character = possible_replacement
                replacing = False

            # automatically is skipped if longer than one letter, so won't enter while replacing
            if (possible_replacement in self.first_letters) and (escaped_until_character == ''):
                if letter_is_escaped:
                    # Escape character. No need to replace. Start over
                    replacing = False
                else:
                    # may be something to replace
                    replacing = True
                    # Handle single character replacements
                    if possible_replacement in self.unique_full_first_letter_replacements:
                        # Got a replacement.
                        # Do replacement
                        output_letter = self.unique_full_first_letter_replacements[possible_replacement]
                        # Reset for next replacement
                        replacing = False

            if replacing:
                # Handle multiple character replacements
                current_replacements = _replacements_starting_with(
                    possible_replacement, self.replacement_dict
                )
                if current_replacements == {}:
                    # No replacements found. Start over
                    replacing = False
                else:
                    # Matching replacements. Determine whether may be another replacement key prefixed by this one
                    if len(current_replacements) == 1:
                        # Single replacement. If it's a full replacement, do it
                        if list(current_replacements.keys())[0] in self.replacement_dict:
                            output_letter = current_replacements[possible_replacement]
                            # Reset for next replacement
                            replacing = False

                        else:
                            # we haven't gotten to the end of this replacement yet, continue until end
                            wipe_replacement = False
                    else:
                        # Multiple matching replacements, continue until only one match
                        wipe_replacement = False

            if wipe_replacement:
                possible_replacement = ''

            previous_letter = letter # Maybe should set to output letter? doesn't matter for current use cases
            output_str += output_letter

        return output_str

    @property
    def first_letters(self) -> List[str]:
        if hasattr(self, '_first_letters'):
            return self._first_letters

        self._first_letters = self._extract_first_letters()
        return self._first_letters

    def _extract_first_letters(self) -> List[str]:
        return _extract_letters_of_index_n(0, self.replacement_dict)

    @property
    def unique_full_first_letter_replacements(self) -> Dict[str, str]:
        if hasattr(self, '_unique_full_first_letter_replacements'):
            return self._unique_full_first_letter_replacements

        self._unique_full_first_letter_replacements = self._extract_unique_full_first_letter_replacements()
        return self._unique_full_first_letter_replacements

    def _extract_unique_full_first_letter_replacements(self) -> Dict[str, str]:
        first_letters_set = set(self.first_letters)
        unique_letters = [letter for letter in first_letters_set if self.first_letters.count(letter) == 1]

        replacements = {}
        for letter in unique_letters:
            replacements.update(_full_replacements_starting_with(letter, self.replacement_dict))

        return replacements


def _full_replacements_starting_with(starts_with_str: str, replacement_dict) -> Dict[str, str]:
    replacements = _replacements_starting_with(starts_with_str, replacement_dict)
    return {key: value for key, value in replacements.items() if len(key) == len(starts_with_str)}

def _replacements_starting_with(starts_with_str: str, replacement_dict) -> Dict[str, str]:
    replacements = replacement_dict.copy()
    for n, letter in enumerate(starts_with_str):
        replacements = _extract_replacements_where_letter_matches_index_n(n, letter, replacements)

    return replacements


def _extract_letters_of_index_n(n: int, dict_: dict) -> List[str]:
    return [key[n] for key in dict_]

def _extract_replacements_where_letter_matches_index_n(n: int, letter: str, dict_: dict) -> Dict[str, str]:
    return {key: value for key, value in dict_.items() if key[n] == letter}


latex_replacer = Replacer(
    replacement_dict,
    escape_next_character_characters=[latex_escape_begin_character],
    escape_until_same_next_character_characters=latex_block_escape_characters
)


from pyexlatex import BibTexArticle


def test_bibtex_eq():
    common_args = (
        'da-engelberg-gao-2011',
        'Zhi Da and Joseph Engelberg and Pengjie Gao',
        'In Search of Attention',
        'Journal of Finance',
        '2011',
    )

    common_kwargs = dict(
        volume='66',
        number='5',
        pages='1461-1499',
        month='10'
    )

    bto1 = BibTexArticle(*common_args, **common_kwargs)
    bto2 = BibTexArticle(*common_args, **common_kwargs)

    assert bto1 == bto2


from typing import Sequence, Any, Tuple
from pyexlatex.presentation.beamer.frame.frame import Frame
from pyexlatex.layouts import CellLayout, GridLayout


class CellFrame(Frame):
    """
    Lay frame out in a grid where each cell is arbitrarily sized
    """

    def __init__(self, content: Sequence[Sequence[Any]], grid_shape: Sequence[Sequence[Tuple[float, float]]], **kwargs):
        """

        Args:
            content:
            grid_shape: list of list of tuples of (width, height), e.g. [
                    [(0.3, 0.4), (0.7, 0.4)],
                    [(0.3, 0.6), (0.7, 0.6)],
                ] would create a 2x2 grid with 40% of height in the first row, 60% in the second row. 30% width in the
                first column, 70% width in the second column.
            **kwargs:
        """
        self.content = content
        self.cells = CellLayout(content, grid_shape)

        super().__init__(self.cells.content, **kwargs)


class GridFrame(Frame):
    """
    Creates a CellFrame, automatically setting widths and heights as an even split based on the shape
    of the content passed
    """

    def __init__(self, content: Sequence[Sequence[Any]], **kwargs):
        self.content = content
        self.grid = GridLayout(content)
        super().__init__(self.grid.content, **kwargs)



from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.models.item import NoOptionsNoContentsItem
from pyexlatex.models.section.base import TextAreaBase


class Centering(NoOptionsNoContentsItem):
    name = 'centering'

    def __init__(self, overlay: Optional['Overlay'] = None):
        super().__init__(self.name, overlay=overlay)


class Center(TextAreaBase):
    """
    Align the passed contents to the center of the container
    """
    name = 'center'

    def __init__(self, content, **kwargs):
        super().__init__(self.name, content, **kwargs)


from pyexlatex.models.item import MultiOptionSimpleItem


class MultiColumnLabel(MultiOptionSimpleItem):
    """
    A label which spans multiple columns in a table
    """
    name = 'multicolumn'

    def __init__(self, contents: str, span: int, align: str='c', add_table_line_break: bool = True):
        self.add_table_line_break = add_table_line_break
        self.span = span
        super().__init__(self.name, span, align, contents)

    def __len__(self):
        return self.span

    def __str__(self) -> str:
        from pyexlatex.models.format.breaks import TableLineBreak
        orig_str = super().__str__()
        if self.add_table_line_break:
            return orig_str + str(TableLineBreak())
        return orig_str




from pyexlatex.presentation.beamer.overlay.offset import Offset
from pyexlatex.presentation.beamer.overlay.overlay_param import OverlayParameter


class NextBase(OverlayParameter):
    next_char = '<invalid, use subclass, not NextBase directly>'

    def __init__(self, offset: int = 0):
        self.offset = Offset(offset)

    def __str__(self) -> str:
        return f'{self.next_char}{self.offset}'


class NextWithIncrement(NextBase):
    """
    Option to be passed to Overlay which makes object appear on the next slide of the frame, and creates an additional
    slide
    """
    next_char = '+'


class NextWithoutIncrement(NextBase):
    """
        Option to be passed to Overlay which makes object appear on the next slide of the frame, and does not creates
        an additional slide
        """
    next_char = '.'



from pyexlatex.models.section.base import ParagraphBase


class SubParagraph(ParagraphBase):
    """
    Part of a paragraph, the smallest section block.
    """
    name = 'subparagraph'


class Paragraph(ParagraphBase):
    """
    A paragraph, the second smallest section block.
    """
    name = 'paragraph'
    next_level_down_class = SubParagraph





from pyexlatex.models.item import SimpleItem


class LaTeX(SimpleItem):
    """
    Displays the string "LaTeX" with special formatting
    """
    name = 'LaTeX'

    def __init__(self, **kwargs):
        super().__init__(self.name, '', **kwargs)



from pyexlatex.models.item import SimpleItem

class Caption(SimpleItem):
    name = 'caption'

    def __init__(self, contents):
        if contents is None:
            contents = ' '
        super().__init__(self.name, contents)



from pyexlatex.models.item import SimpleItem


class VSpace(SimpleItem):
    """
    Manually insert vertical spacing, of a given height.
    """
    name = 'vspace'

    def __init__(self, height: float = 0.5, units: str = 'cm'):
        self.height = height
        self.units = units
        super().__init__(self.name, contents=f'{height}{units}')



from mixins.repr import ReprMixin
from pyexlatex.presentation.beamer.overlay.overlay_param import OverlayParameter


class Range(ReprMixin, OverlayParameter):
    """
    Option to be passed to Overlay which says object should exist on the frame on a certain range of slides
    e.g. slide 1-3 of the frame.
    """
    repr_cols = ['low', 'high']

    def __init__(self, low: int, high: int):
        self.low = low
        self.high = high

    def __str__(self) -> str:
        return f'{self.low}-{self.high}'


from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.models.item import SimpleItem


class Begin(SimpleItem):
    name = 'begin'

    def __init__(self, env: str, modifiers: Optional[str] = None, overlay: Optional['Overlay'] = None):
        self.env = env
        super().__init__(self.name, env, modifiers, overlay=overlay)



import copy
from typing import List, Sequence

from pyexlatex.table.models.panels.grid import GridShape
from pyexlatex.table.models.labels.table import LabelTable
from pyexlatex.table.models.labels.collection import LabelCollection
from pyexlatex.table.models.table.section import TableSection

def common_column_labels(grid: GridShape, use_object_equality=True, enforce_label_order=True):
    axis = 1 # columns
    all_column_ints = list(range(grid.shape[1]))

    return _selected_common_labels_for_axis(
        grid,
        selections=all_column_ints,
        axis=axis,
        use_object_equality=use_object_equality,
        enforce_label_order=enforce_label_order
    )

def common_row_labels(grid: GridShape, use_object_equality=True, enforce_label_order=True):
    axis = 0  # rows
    all_row_ints = list(range(grid.shape[0]))

    return _selected_common_labels_for_axis(
        grid,
        selections=all_row_ints,
        axis=axis,
        use_object_equality=use_object_equality,
        enforce_label_order=enforce_label_order
    )


def _selected_common_labels_for_axis(grid: GridShape, selections: Sequence[int]=(0,), axis: int=0, use_object_equality=True,
                                     enforce_label_order=True):
    common_label_tables: List[LabelTable] = []
    for i in selections:
        common_label_tables.append(
            _common_labels(
                grid,
                i,
                axis=axis,
                use_object_equality=use_object_equality,
                enforce_label_order=enforce_label_order
            )
        )

    non_none_tables = [table for table in common_label_tables if table is not None]

    if non_none_tables == []:
        return None

    return non_none_tables


def _common_labels(grid: GridShape, num: int, axis: int=0, use_object_equality=True,
                   enforce_label_order=True):
    subgrid = _get_subgrid(
        grid=grid,
        num=num,
        axis=axis
    )

    label_attr = _get_label_attr(axis=axis)

    label_tables: List[LabelTable] = [getattr(section, label_attr, None) for section in subgrid]

    # first labels missing, no consolidation to be done, consolidated labels are None
    if label_tables[0] is None:
        return None

    common_label_table = LabelTable([])
    for i, label_collection in enumerate(label_tables[0]):
        stored_match = False # only want to add each matched collection once. use boolean to track
        for label_table in label_tables[1:]:
            # If there is a corresponding label table and it has this index label collection
            if label_table is not None and i < len(label_table.label_collections):
                match = _compare_label_collections(
                    label_collection,
                    label_table[i],
                    use_object_equality=use_object_equality
                )
            else:
                # No labels, nothing to consolidate
                match = False
            if match:
                if not stored_match:
                    common_label_table.append(label_collection)
                    stored_match = True
            else:
                if enforce_label_order:
                    break # as soon as one label collection doesn't match, stop consolidating
                else:
                    continue # don't worry about non-match, continue consolidating

    if common_label_table.is_empty:
        return None

    return common_label_table

def remove_label_collections_from_grid(grid: GridShape, column_labels: List[LabelTable] = None,
                                       row_labels: List[LabelTable] = None, use_object_equality=True):
    for row in grid:
        for section in row:
            if column_labels is not None:
                for label_table in column_labels:
                    _remove_label_collections(
                        section,
                        label_table,
                        axis=1,
                        use_object_equality=use_object_equality,
                        inplace=True
                    )
            if row_labels is not None:
                for label_table in row_labels:
                    _remove_label_collections(
                        section,
                        label_table,
                        axis=0,
                        use_object_equality=use_object_equality,
                        inplace=True
                    )

def _remove_label_collections(section: TableSection, label_table: LabelTable, axis: int=0,
                              use_object_equality=True, inplace=False):

    # Handle section not having labels for this axis
    label_attr = _get_label_attr(axis=axis)
    if not hasattr(section, label_attr):
        return section

    # Now has labels for this axis. Create a copy to avoid modifying original
    if not inplace:
        section = copy.deepcopy(section)

    for label_collection in label_table:
        section_label_table: LabelTable = getattr(section, label_attr, [])
        if section_label_table is not None:
            for section_label_collection in section_label_table:
                match = _compare_label_collections(
                    label_collection,
                    section_label_collection,
                    use_object_equality=use_object_equality
                )
                if match:
                    section_label_table.remove(section_label_collection)


    # once all label collections have been removed, remove table
    section_label_table = getattr(section, label_attr, False)
    if section_label_table and section_label_table.label_collections == []:
        setattr(section, label_attr, None)

    return section


def _get_label_attr(axis: int=0):
    # select rows
    if axis == 0:
        return 'row_labels'
    # select columns
    elif axis == 1:
        return 'column_labels'
    else:
        raise ValueError(f'axis must be 0 or 1, got {axis}')

def _get_subgrid(grid: GridShape, num: int, axis: int=0):
    # select rows
    if axis == 0:
        return _grid_if_not(grid[num])
    # select columns
    elif axis == 1:
        return _grid_if_not(grid[:, num])
    else:
        raise ValueError(f'axis must be 0 or 1, got {axis}')

def _grid_if_not(ambiguous_grid):
    if isinstance(ambiguous_grid, GridShape):
        return ambiguous_grid
    else:
        return GridShape([ambiguous_grid])

def _compare_label_collections(collection1: LabelCollection, collection2: LabelCollection, use_object_equality=True):
    if use_object_equality:
        return collection1 == collection2
    else:
        return collection1.matches(collection2)

def _add_to_label_table_if_not_in_label_table(label_table: LabelTable, label_collection: LabelCollection,
                                              use_object_equality=True):
    """
    Note: inplace
    """
    # don't want to keep adding match over and over. need to check if match is already
    # stored in the common label table. must check two different ways depending on whether
    # we are using object equality or string consolidation
    if use_object_equality and label_collection not in label_table:
        label_table.append(label_collection)
    if (not use_object_equality) and (not label_table.contains(label_collection)):
        label_table.append(label_collection)


from pyexlatex.models.item import NoOptionsNoContentsItem


class Justifying(NoOptionsNoContentsItem):
    """
    Returns paragraph to the default of justifying text on both left and right. Useful if paragraph is currently
    ragged or sloppy, which occurs when using NoLineBreak.
    """
    name = 'justifying'

    def __init__(self, **kwargs):
        self.add_package('ragged2e')
        super().__init__(self.name, **kwargs)


from typing import Optional, List, Dict, Any
import warnings
from data import Data
import latex
from latex.exc import LatexBuildError
from pyexlatex.logic.pdf.errors.exc import (
    TooManyUnprocessedFloatsException,
    OutputLoopConsecutiveDeadCycles,
    LatexException,
    exception_manager
)
from pyexlatex.logic.pdf.api.exc_handler.main import APIExceptionHandler
from pyexlatex.logic.pdf.api.exc_handler.prepend.typing import PrependKwargsDict, PrependItemsDict
from pyexlatex.logic.pdf.api.exc_handler.prepend.main import add_prepend_items_dict_to_latex_str
from pyexlatex.logic.pdf.api.builders.lualatex import LuaLatexBuilder

def latex_str_to_pdf_obj(latex_str: str, texinputs: Optional[List[str]] = None, run_bibtex: bool = False,
                         retries_remaining: int = 3,
                         prepend_items_dict: PrependItemsDict = None,
                         prepend_kwargs_dict: PrependKwargsDict = None) -> Data:
    try:
        new_latex_str = add_prepend_items_dict_to_latex_str(prepend_items_dict, latex_str)
        return _latex_to_pdf_obj(new_latex_str, texinputs=texinputs, run_bibtex=run_bibtex)
    except LatexBuildError as e:
        exceptions = exception_manager.exceptions_from_latex_build_error(e)
        handler = APIExceptionHandler(
            exceptions,
            e,
            latex_str,
            prepend_kwargs_dict=prepend_kwargs_dict,
            prepend_items_dict=prepend_items_dict,
            retries_remaining=retries_remaining,
            texinputs=texinputs,
            run_bibtex=run_bibtex
        )
        return handler.handle_exceptions()


def _latex_to_pdf_obj(latex_str: str, texinputs: Optional[List[str]] = None,
                      run_bibtex: bool = False) -> Data:
    if texinputs is None:
        texinputs = []
    pdf = LuaLatexBuilder().build_pdf(latex_str, texinputs=texinputs, run_bibtex=run_bibtex)
    return pdf




from pyexlatex.models.item import Item
from pyexlatex.models.package import Package
from pyexlatex.constants.flags import HAS_MINTED


class Minted(Item):
    name = 'minted'

    def __init__(self, content, language: str):
        self.init_data()
        self.data.packages.append(Package('minted'))
        self.data.flags.append(HAS_MINTED)
        super().__init__(self.name, content, env_modifiers=self._wrap_with_braces(language))



from typing import Iterable


def _is_collection(collection, desired_obj_types):
    if isinstance(collection, str):
        return False
    if isinstance(collection, desired_obj_types):
        return False
    return isinstance(collection, Iterable)


from typing import Union, TYPE_CHECKING
from pyexlatex.models.item import MultiOptionSimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.texgen import _multi_option_item_str
if TYPE_CHECKING:
    from pyexlatex.models.format.text.color.deftypes.base import ColorDefinition


class TextColor(TextAreaMixin, MultiOptionSimpleItem):
    """
    Changes the color of text.

    Usage:
    >>> import pyexlatex as pl
    >>> pl.TextColor('something', pl.Hex('59bf37'))
    >>> pl.TextColor('something', pl.RGB(89, 191, 55))
    >>> pl.TextColor('something', 'blue', opacity=0.5)
    """
    name = 'textcolor'

    def __init__(self, content, color: Union[str, 'ColorDefinition'], opacity: float = 1.0, **kwargs):
        """

        :param content:
        :param color: can pass a string name of an existing color, such as 'green' or 'red', or can pass a defined
            color, using one of the ColorDefinition models, e.g. RGB or Hex
        :param opacity:
        :param kwargs:
        """
        self.opacity = opacity
        self.init_data()
        self.data.packages.append('xcolor')
        self.color = self.process_color_get_name(color)
        self.content = content
        contents = self.format_contents(content)
        MultiOptionSimpleItem.__init__(self, self.name, self.color_str, contents, **kwargs)

    @property
    def opacity_str(self) -> str:
        if self.opacity != 1:
            return f'!{self.opacity * 100:.0f}'

        return ''

    @property
    def color_str(self) -> str:
        return f'{self.color}{self.opacity_str}'

    def process_color_get_name(self, color: Union[str, 'ColorDefinition']) -> str:
        if isinstance(color, str):
            return color

        # Got color definition. Need to add to preamble and reference by the name of the color
        self.data.packages.append(color)
        return color.color_name

    def __str__(self):
        return _multi_option_item_str(self.name, self.color_str, self.content, overlay=self.overlay)


from pyexlatex.models.mixins import IsSpecificClassMixin
from pyexlatex.models.item import IsLatexItemMixin


class DocumentItem(IsSpecificClassMixin, IsLatexItemMixin):
    """
    Used for differentiating which items can be directly included in a document
    """
    is_DocumentItem = True



import re
from typing import Tuple

from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models.item import Item
from pyexlatex.models.caption import Caption
from pyexlatex.models.label import Label
from pyexlatex.logic.builder import build_figure_content
from pyexlatex.models.containeritem import ContainerItem

POSITION_STR_PATTERN = re.compile(r'\[([tcb])\]{([\d.\w\\]+)}')


class Subfigure(ContainerItem, Item):
    """
    Used for more control over building figures
    """
    name = 'subfigure'

    def __init__(self, filepath, caption=None, label=None, centering=True, position_str=r'[t]{0.45\linewidth}',
                 cache: bool = True):
        self.graphic = Graphic(filepath, cache=cache)
        self.caption = Caption(caption) if caption else None
        self.label = Label(label) if label else None
        self.position_str = position_str
        self.anchor, self.width = anchor_and_width_from_position_str(self.position_str)

        self.add_data_from_content(self.graphic)

        contents = build_figure_content(
            [self.graphic],
            caption=self.caption,
            label=self.label,
            centering=centering,
            position_str=position_str
        )

        super().__init__(self.name, contents)

    def __repr__(self):
        return f'<Subfigure({self.graphic.filepath}, caption={self.caption})>'


def anchor_and_width_from_position_str(position: str) -> Tuple[str, str]:
    if not isinstance(position, str):
        position = str(position)
    match = re.match(POSITION_STR_PATTERN, position)
    if not match:
        raise ValueError(f'could not parse position {position}')
    
    anchor = match.group(1)
    width = match.group(2)
    return anchor, width






from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.models.item import SimpleItem


class Uncover(SimpleItem):
    name = 'uncover'

    def __init__(self, contents, overlay: Optional['Overlay'] = None):
        super().__init__(self.name, contents, overlay=overlay)


import os
import subprocess
from subprocess import CalledProcessError

from future.utils import raise_from
from data import Data as I
from data.decorators import data
from shutilwhich import which
from tempdir import TempDir

from latex.exc import LatexBuildError


class LuaLatexBuilder:
    """A simple lualatex based buidler for LaTeX files.

    Builds LaTeX files by copying them to a temporary directly and running
    ``lualatex`` until the associated ``.aux`` file stops changing.

    .. note:: This may miss changes if ``biblatex`` or other additional tools
              are used. Usually, the :class:`~latex.build.LatexMkBuilder` will
              give more reliable results.

    :param lualatex: The path to the ``pdflatex`` binary (will looked up on
                    ``$PATH``).
    :param max_runs: An integer providing an upper limit on the amount of times
                     ``pdflatex`` can be rerun before an exception is thrown.
    """

    def __init__(self, lualatex='lualatex', bibtex: str = 'bibtex', max_runs=15):
        self.lualatex = lualatex
        self.bibtex = bibtex
        self.max_runs = 15

    @data('source')
    def build_pdf(self, source, texinputs=[], run_bibtex: bool = False):
        with TempDir() as tmpdir,\
                source.temp_saved(suffix='.latex', dir=tmpdir) as tmp:

            # close temp file, so other processes can access it also on Windows
            tmp.close()
            called_bibtex = False

            # calculate output filename
            base_fn = os.path.splitext(tmp.name)[0]
            output_fn = base_fn + '.pdf'
            aux_fn = base_fn + '.aux'
            args = [self.lualatex, '-interaction=batchmode',
                    '-halt-on-error',
                    '-shell-escape', '-file-line-error',
                    tmp.name]

            # create environment
            newenv = os.environ.copy()
            inputs_value = os.pathsep.join(texinputs) + os.pathsep
            newenv['TEXINPUTS'] = inputs_value
            newenv['BSTINPUTS'] = inputs_value

            # run until aux file settles
            prev_aux = None
            runs_left = self.max_runs
            while runs_left:
                try:
                    subprocess.check_call(args,
                                          cwd=tmpdir,
                                          env=newenv,
                                          stdin=open(os.devnull, 'r'),
                                          stdout=open(os.devnull, 'w'), )
                except CalledProcessError as e:
                    raise_from(LatexBuildError(base_fn + '.log'), e)

                # check aux-file
                aux = open(aux_fn, 'rb').read()

                if aux == prev_aux:
                    # Stable aux file
                    if run_bibtex and not called_bibtex:
                        called_bibtex = True  # ensure only called once
                        bibtex_args = [self.bibtex, os.path.basename(aux_fn)]
                        try:
                            subprocess.check_call(bibtex_args,
                                              cwd=tmpdir,
                                              env=newenv,
                                              stdin=open(os.devnull, 'r'),
                                              stdout=open(os.devnull, 'w'), )
                        except CalledProcessError as e:
                            # TODO [#9]: better handling for LaTeX exceptions
                            #
                            # Parse log file, raise proper exception
                            with open(base_fn + '.blg', 'r') as f:
                                log_contents = f.read()
                            raise Exception(log_contents)
                        continue  # go back into the loop to process with biblography
                    break

                prev_aux = aux
                runs_left -= 1
            else:
                raise RuntimeError(
                    'Maximum number of runs ({}) without a stable .aux file '
                    'reached.'.format(self.max_runs))

            return I(open(output_fn, 'rb').read(), encoding=None)

    def is_available(self):
        return bool(which(self.lualatex))



from pyexlatex.models.item import Item
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.commands.newenvironment import NewEnvironment


class EnvironmentTemplate(TextAreaMixin, Item):
    """
    Base class which makes it easy to create new environment types.

    Examples:
    >>> import pyexlatex as pl
    >>> class BoldEnvironment(pl.EnvironmentTemplate):
    >>>     name = 'myenv'
    >>>     begin_def = r'\bf'
    >>>     end_def = ''
    >>>
    >>> print(pl.Document(BoldEnvironment('woo')))
    \documentclass[]{article}
    \newenvironment{myenv}{\bf}{}
    ...
    \begin{document}
    \begin{myenv}
    woo
    \end{myenv}
    \end{document}


    """
    begin_def = ''
    end_def = ''

    def __init__(self, contents, **kwargs):
        self.init_data()
        self.data.begin_document_items.append(self._environment_def)
        super().__init__(self.name, contents, **kwargs)

    @property
    def _environment_def(self) -> NewEnvironment:
        return NewEnvironment(
            self.name,
            self.begin_def,
            self.end_def
        )



class StringAdditionMixin:

    def __add__(self, other):
        return str(self) + str(other)

    def __radd__(self, other):
        return str(other) + str(self)

    def join(self, iterable):
        return self.__str__().join(str(i) for i in iterable)


class StringEqMixin:

    def __eq__(self, value):
        return str(self) == str(value)

    def __hash__(self):
        return hash(str(self))


class IsSpecificClassMixin:
    """
    creates attribute is_ClassName (with whatever class name)

    be sure to put this mixin first among multiple classes
    """

    def __init__(self, *args, **kwargs):
        class_name = self.__class__.__name__
        setattr(self, f'is_{class_name}', True)
        super().__init__(*args, **kwargs)



from typing import Dict, List
from pyexlatex.models.commands.renewcommand import ReNewCommand


class SectionNumberingFormatter(ReNewCommand):
    """
    applies formatting for section numbers
    """

    def __init__(self, section_type: str, section_format: str):
        section_type_format_command_name = f'the{section_type}'  # e.g. \thesection \thesubsection
        super().__init__(section_type_format_command_name, section_format)

    @classmethod
    def list_from_string_format_dict(cls, string_format_dict: Dict[str, str]) -> List['SectionNumberingFormatter']:
        """

        Args:
            string_format_dict: e.g. dict(
                section=r'\Roman{section}',
                subsection=r'\Alph{subsection}'
            )

        Returns:

        """
        formatters = []
        for section_type, section_format in string_format_dict.items():
            formatters.append(
                cls(section_type, section_format)
            )
        return formatters




from typing import List, Optional, Sequence, Union, TYPE_CHECKING, Callable

from pyexlatex.presentation.beamer.control.navigation import AddNavigationHeader
from pyexlatex.presentation.beamer.templates.control.tocsection import TableOfContentsAtBeginSection

if TYPE_CHECKING:
    from pyexlatex.models.documentsetup import DocumentSetupData
from copy import deepcopy
from pyexlatex.models.document import DocumentBase
from pyexlatex.typing import ItemOrListOfItems
from pyexlatex.models.package import Package
from pyexlatex.models.control.documentclass.documentclass import DocumentClass
from pyexlatex.models.control.mode import Mode
from pyexlatex.presentation.beamer.theme.usetheme import UseTheme
from pyexlatex.models.title.frame import TitleFrame, should_create_title_frame
from pyexlatex.models.item import ItemBase, Item
from pyexlatex.presentation.beamer.templates.lists.dim_reveal_items import eliminate_dim_reveal
from pyexlatex.exc import NoPackageWithNameException


class Presentation(DocumentBase):
    """
    The main high-level class for creating presentations, can render presentations as PDF in either a presenter
    style or a handouts style.
    """
    name = 'document'

    def __init__(self, content: ItemOrListOfItems, packages: List[Package]=None,
                 pre_env_contents: Optional[ItemOrListOfItems] = None,
                 title: Optional[str] = None, authors: Optional[Union[str, Sequence[str]]] = None,
                 date: Optional[str] = None,
                 short_title: Optional[str] = None, subtitle: Optional[str] = None, short_author: Optional[str] = None,
                 institutions: Optional[Sequence[Sequence[str]]] = None, short_institution: Optional[str] = None,
                 font_size: Optional[float] = 11, theme: str = 'Madrid', backend: str = 'beamer',
                 nav_header: bool = False, toc_sections: bool = False, handouts: bool = False,
                 pre_output_func: Optional[Callable] = None):

        self.init_data()
        self.title_frame = None

        from pyexlatex.models.documentitem import DocumentItem
        content_list: List[Union[Item, ItemBase]]
        if isinstance(content, (ItemBase, str)):
            content_list = [content]
        else:
            # don't overwrite existing content
            content_list = deepcopy(content)  # type: ignore

        pre_env_contents_list: List[Union[Item, ItemBase]]
        if pre_env_contents is None:
            pre_env_contents_list = []
        elif isinstance(pre_env_contents, (ItemBase, str)):
            pre_env_contents_list = [pre_env_contents]
        else:
            pre_env_contents_list = pre_env_contents  # type: ignore

        if backend != 'beamer':
            raise NotImplementedError('only beamer backend is currently supported for Presentation')

        doc_class_options: Optional[List[str]]
        if handouts:
            doc_class_options = ['handout']
            elimate_overlays(content_list)
            eliminate_dim_reveal(content_list)
        else:
            doc_class_options = None

        self.document_class_obj = DocumentClass(
            document_type=backend,
            font_size=font_size,
            options=doc_class_options
        )

        data_cleanup_func: Optional[Callable[[DocumentSetupData], None]]
        if backend == 'beamer':
            # Remove hyperref package if added, as may conflict with beamer which already loads it
            def remove_hyperref(data: 'DocumentSetupData') -> None:
                try:
                    data.packages.delete_by_name('hyperref')
                except NoPackageWithNameException:
                    pass
            data_cleanup_func = remove_hyperref

            # TODO [#17]: add support for custom presentation theming, not just passing main theme str
            styles = [UseTheme(theme)]
            self.data.packages.append(
                Mode(
                    mode_type='presentation',
                    contents=styles
                )
            )

            if should_create_title_frame(title, authors, date, subtitle, institutions):
                self.title_frame = TitleFrame(
                    title=title,
                    authors=authors,
                    date=date,
                    short_title=short_title,
                    subtitle=subtitle,
                    short_author=short_author,
                    institutions=institutions,
                    short_institution=short_institution
                )
                content_list.insert(0, self.title_frame)

            if nav_header:
                pre_env_contents_list.append(AddNavigationHeader())

            if toc_sections:
                pre_env_contents_list.append(TableOfContentsAtBeginSection())

        else:
            data_cleanup_func = None

        super().__init__(
            content_list,
            packages=packages,
            pre_env_contents=pre_env_contents_list,
            data_cleanup_func=data_cleanup_func,
            pre_output_func=pre_output_func
        )


def elimate_overlays(content):
    """
    Eliminates overlays from nested content. Modifies in place
    """
    if hasattr(content, 'content'):
        elimate_overlays(content.content)
    if hasattr(content, 'contents'):
        elimate_overlays(content.contents)
    if hasattr(content, 'overlay'):
        content.overlay = None
    if isinstance(content, (list, tuple)):
        [elimate_overlays(c) for c in content]

